library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
library unisim;
use unisim.vcomponents.all;

--This module digitizes four channels at once. Four inputs come from four different ADCs. 
--Datasheet for ADCs: http://www.ti.com/lit/ds/symlink/ads7886.pdf  see page 6 for timing
--The ADCs output information from MSB to LSB, so the values are filled in starting
--from the 15th bit to the 0th bit. The four MSBs are zero, and bits 11 - 0 represent
--the value being digitized.

--The incoming clock runs at 60MHz, and the ADC runs at 15MHz(actually up to 20MHz). 
--The clk_div signal is used to translate 60MHz down to 15MHz. The entire cycle
--lasts for 17 clock cycles at 15MHz, or 68 clock cycles at 60MHz

entity adc is
  port (
    clk              : in  std_logic;
    adc_pin_1        : in  std_logic; 
    adc_pin_2        : in  std_logic;
    adc_pin_3        : in  std_logic;
    adc_pin_4        : in  std_logic;
    cs               : out std_logic; -- control signal going to ADC
    adc_word_final_1 : out std_logic_vector(15 downto 0);
    adc_word_final_2 : out std_logic_vector(15 downto 0);
    adc_word_final_3 : out std_logic_vector(15 downto 0);
    adc_word_final_4 : out std_logic_vector(15 downto 0);
    adc_clk          : out std_logic; -- 15MHz output clock to run ADCs
    ram_wr_en        : out std_logic; -- a key output that alerts the other modules every time a full word is constructed
    rst              : in  std_logic); 

end adc;

architecture Behavioral of adc is

  signal cnt_sclk         : std_logic_vector(4 downto 0);
  signal adc_word_build_1 : std_logic_vector(15 downto 0);
  signal adc_word_build_2 : std_logic_vector(15 downto 0);
  signal adc_word_build_3 : std_logic_vector(15 downto 0);
  signal adc_word_build_4 : std_logic_vector(15 downto 0);
  signal clk_div          : std_logic_vector(1 downto 0);



begin

  adc_read : process(clk, rst)
  begin
    adc_word_build_1(15) <= '0';

    if rst = '1' then
      cs               <= '1';
      cnt_sclk         <= "00000";
      clk_div          <= "00";
      ram_wr_en        <= '0';
      adc_word_build_1 <= (others => '0');
      adc_word_build_2 <= (others => '0');
      adc_word_build_3 <= (others => '0');
      adc_word_build_4 <= (others => '0');
    elsif (clk'event and clk = '1') then
      clk_div <= clk_div + 1;
      if (clk_div = "00") then           -- 15MHz clock going to ADC
        adc_clk   <= '0';
        ram_wr_en <= '0';
        cnt_sclk  <= cnt_sclk + 1;
      elsif (clk_div = "10") then
        adc_clk <= '1';
        if (cnt_sclk = "00000") then
          cs        <= '0';
          ram_wr_en <= '0';
        elsif (cnt_sclk = "10000") then  --quiet time; ram_wr_enable signals that the words are ready
          cs               <= '1';
          adc_word_final_1 <= adc_word_build_1;
          adc_word_final_2 <= adc_word_build_2;
          adc_word_final_3 <= adc_word_build_3;
          adc_word_final_4 <= adc_word_build_4;
          ram_wr_en        <= '1'; 
          cnt_sclk         <= "11111";
        else  --build the words one bit at a time
          adc_word_build_1(15-conv_integer(cnt_sclk)) <= adc_pin_1;
          adc_word_build_2(15-conv_integer(cnt_sclk)) <= adc_pin_2;
          adc_word_build_3(15-conv_integer(cnt_sclk)) <= adc_pin_3;
          adc_word_build_4(15-conv_integer(cnt_sclk)) <= adc_pin_4;
          ram_wr_en                                   <= '0';
          cs                                          <= '0';
        end if;
      else
        ram_wr_en <= '0';
      end if;
    end if;
  end process;
end Behavioral;
