-- Storage module. Current design uses block ram for storage,
-- but code is written to support using an external ram. 
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity storage is
  port (
    clk       : in std_logic;
    rst       : in std_logic;
	 --signal from ADC saying that new sample is available
    ram_wr_en : in std_logic;

	 --threshold crossing signals from threshold modules
    thresh_signal_1 : in std_logic;
    thresh_signal_2 : in std_logic;
    thresh_signal_3 : in std_logic;
    thresh_signal_4 : in std_logic;

	 --words coming from block ram buffer
    ram_word_1 : in std_logic_vector(15 downto 0);
    ram_word_2 : in std_logic_vector(15 downto 0);
    ram_word_3 : in std_logic_vector(15 downto 0);
    ram_word_4 : in std_logic_vector(15 downto 0);

	 --ram full flags sent to 
    ram_full_1 : out std_logic;
    ram_full_2 : out std_logic;
    ram_full_3 : out std_logic;
    ram_full_4 : out std_logic;

	 --current writing address
	 wr_addr_sto_out : out std_logic_vector(15 downto 0);
	 --current reading/transmitting address
	 out_addr_out : out std_logic_vector(15 downto 0);

	 --reset flags from tx_rx
	 tx_reset_1   : in std_logic;
	 tx_reset_2   : in std_logic;
	 tx_reset_3   : in std_logic;
	 tx_reset_4   : in std_logic;

	 --control signals from USB module
	 txe_lo : in std_logic;
	 wr_lo : in std_logic;
	 
	 --wave count going to tx_rx for status byte
    wave_cnt_1 : out std_logic_vector(1 downto 0);
    wave_cnt_2 : out std_logic_vector(1 downto 0);
    wave_cnt_3 : out std_logic_vector(1 downto 0);
    wave_cnt_4 : out std_logic_vector(1 downto 0);
	 
	 --control signals for transmitting from storage module and tx_rx
	 read_begin : in std_logic;
	 read_pause : out std_logic;
	 read_end	: out std_logic;
	
	 --beginning and ending address for transmitting
    final_word_addr : in std_logic_vector(15 downto 0);  --actual will have more bits
	 final_word_addr_end : in std_logic_vector(15 downto 0);
	 --final output word (sent one byte at a time) to tx_rx
    final_word_1 : out std_logic_vector(7 downto 0));

end storage;

architecture syn of storage is
  --storage ram that saves data coming from the block ram buffers
  --4 channels that all have an equal amount of space...2048 bytes,
  --1024 words, meaning 4 waveforms per channel
  type ram_type_5 is array (8191 downto 0) of std_logic_vector (7 downto 0);
  signal ram : ram_type_5;

  signal storage_state : std_logic;
  signal ram_word : std_logic_vector(15 downto 0);

  --addresses that hold the different channels and pass them on to wr_addr_sto
  signal wr_addr_sto_1 : std_logic_vector(15 downto 0);
  signal wr_addr_sto_2 : std_logic_vector(15 downto 0);
  signal wr_addr_sto_3 : std_logic_vector(15 downto 0);
  signal wr_addr_sto_4 : std_logic_vector(15 downto 0);
  signal wr_addr_sto   : std_logic_vector(15 downto 0);
  
  --holds 
  signal wr_addr_all : std_logic_vector(15 downto 0); 
  
  --signals that hold current and next destination for writing to storage
  --ram based on current position and thresh_signal flags
  signal r: std_logic_vector(2 downto 0);
  signal a: std_logic_vector(2 downto 0); 

  signal read_state: std_logic_vector(2 downto 0);
  
  --signals that are assigned initial values based
  --on starting and finishing addresses from tx_rx
  signal out_addr: std_logic_vector(15 downto 0);
  signal out_addr_end: std_logic_vector(15 downto 0);
  
  --timing signals for writing from storage to tx_tx
  signal sto_delay: std_logic_vector(1 downto 0);
  signal delay_begin: std_logic;
  signal wr_delay: std_logic;
  
--  signal mult_sel_2: std_logic_vector(2 downto 0);
  signal ram_state: std_logic;

  
begin

  wave_cnt_1 <= wr_addr_sto_1(10 downto 9);
  wave_cnt_2 <= wr_addr_sto_2(10 downto 9);
  wave_cnt_3 <= wr_addr_sto_3(10 downto 9);
  wave_cnt_4 <= wr_addr_sto_4(10 downto 9);
  
  --ridiculously long logic equations. this is used to figure out where to go next. 
  --at any point this module is either doing nothing, writing from the ram buffer into the storage ram
  --or sending data from the storage ram to the tx_rx module. this module is written to allow 
  --any of this to happen dynamically. so if this module is in the middle of writing to tx_tx
  -- and three channels need to be written into storage ram, it will interrupt just long enough 
  --to write the samples into storage ram, then continue transmitting
  --if nothing needs to be written into storage ram, it can transmit without being interrupted
  --"a" is a three bit signal that is the current location
  --000 channel 1
  --001 channel 2
  --010 channel 3
  --011 channel 4
  --100 ready to transmit or transmitting
  --"r" is the next destination, depending on current location and the values of all the thresh signals, 
  r(2) <= (a(1) and a(0)) or (a(1) and not(thresh_signal_4)) or (a(0) and not(thresh_signal_3) and not(thresh_signal_4)) or (not(a(2)) and not(thresh_signal_2) and not(thresh_signal_3) and not(thresh_signal_4)) or (not(thresh_signal_1) and not(thresh_signal_2) and not(thresh_signal_3) and not(thresh_signal_4));
  r(1) <= (a(1) and not(a(0)) and thresh_signal_4) or (not(a(2)) and not(a(1)) and not(thresh_signal_2) and thresh_signal_3) or (not(a(1)) and a(0) and not(thresh_signal_3) and thresh_signal_4) or (not(a(1)) and a(0) and thresh_signal_3) or (not(a(2)) and not(a(0)) and not(thresh_signal_2) and not(thresh_signal_3) and thresh_signal_4) or (not(a(0)) and not(thresh_signal_1) and not(thresh_signal_2) and not(thresh_signal_3) and thresh_signal_4) or (not(a(1)) and not(thresh_signal_1) and not(thresh_signal_2) and thresh_signal_3);
  r(0) <= (a(1) and not(a(0)) and thresh_signal_4) or (not(a(2)) and not(a(1)) and not(a(0)) and thresh_signal_2) or (not(a(1)) and a(0) and not(thresh_signal_3) and thresh_signal_4) or (not(a(2)) and not(a(0)) and not(thresh_signal_2) and not(thresh_signal_3) and thresh_signal_4) or (not(a(0)) and not(thresh_signal_1) and  not(thresh_signal_2) and not(thresh_signal_3) and thresh_signal_4) or (not(a(1)) and not(a(0)) and not(thresh_signal_1) and thresh_signal_2);   
	

  wr_addr_sto_out <= wr_addr_sto;
  out_addr_out <= out_addr;
--  mult_sel_2 <= a;
	 
  
	 --multiplexer sets writing address
	 with r select               
		wr_addr_sto <=
		wr_addr_sto_1  when "000",       
		wr_addr_sto_2  when "001",
		wr_addr_sto_3  when "010",
    	wr_addr_sto_4  when "011",
		x"0000"          when others; 
		
	 --multiplexer assigns word from proper block ram buffer channel
	 with a select 
		ram_word <=
		ram_word_1  when "000",       
		ram_word_2  when "001",
		ram_word_3  when "010",
    	ram_word_4  when "011",
		x"0000"          when others; 

  process (clk, rst)
  begin

    if rst = '1' then
      storage_state  <= '1';
      wr_addr_sto_1  <= (others => '0');
      wr_addr_sto_2  <= x"0800";
      wr_addr_sto_3  <= x"1000";
      wr_addr_sto_4  <= x"1800";
      wr_addr_all    <= (others => '0');
		out_addr			<= (others => '0');
		out_addr_end	<= (others => '0');
      ram_full_1     <= '0';
      ram_full_2     <= '0';
      ram_full_3     <= '0';
      ram_full_4     <= '0';
		read_pause		<= '0';
		read_end			<= '0';
		read_state		<= (others => '0');
		sto_delay		<= (others => '0');
		delay_begin		<= '0';
		a					<= "100";
		ram_state		<= '1';
		wr_delay <= '0';
	elsif (clk'event and clk = '1') then
    
		--individual channel resets
		if tx_reset_1 = '1' then
			ram_full_1 <= '0';
			wr_addr_sto_1 <= (others => '0');
		end if;
		
		if tx_reset_2 = '1' then
			ram_full_2 <= '0';
			wr_addr_sto_2 <= x"0800";
		end if;
		
      if tx_reset_3 = '1' then
			ram_full_3 <= '0';
			wr_addr_sto_3 <= x"1000";
		end if;
		
		if tx_reset_4 = '1' then
			ram_full_4 <= '0';
			wr_addr_sto_4 <= x"1800";
		end if;
      	 
		case storage_state is
		  when '0' =>
			--this means the module is in transmit ready state
			 if a = "100" then
				storage_state <= '1';
			 else		
			 --if not in transmit ready state, write from selected block ram buffers into storage ram
				case ram_state is
					when '0' =>
					  --writing of MSB of sample, the value of wr_addr_all and ram_word are controlled by
					  --"a" and "r" and the multiplexers
					  ram(conv_integer(wr_addr_all)) <= ram_word(15 downto 8);
					  ram_state <='1';
					  --assign next address
					  wr_addr_all <= wr_addr_sto;
					  --assign next destination
					  a <= r;
					  case a is
						 when "000" =>
							--individual ram full flags
							--channel 1
							if wr_addr_sto_1 = x"01FE" then
								ram_full_1 <= '1';
							else
								wr_addr_sto_1 <= wr_addr_sto_1 + 2;
							end if;
							--channel 2
						 when "001" =>
							if (wr_addr_sto_2 = x"09FE") then
								ram_full_2 <= '1';
							else
								wr_addr_sto_2 <= wr_addr_sto_2 + 2;
							end if;
							--channel 3
						 when "010" =>
							if wr_addr_sto_3 = x"11FE" then
								ram_full_3 <= '1';
							else
								wr_addr_sto_3 <= wr_addr_sto_3 + 2;
							end if;
							--channel 4
						 when "011" =>
							if wr_addr_sto_4 = x"19FE" then
								ram_full_4 <= '1';
							else
								wr_addr_sto_4 <= wr_addr_sto_4 + 2;
							end if;
						 when others =>
					  end case;
				  when others =>
					  --writing of LSB of sample, the value of wr_addr_all and ram_word are controlled by
					  --"a" and "r" and the multiplexers
					  ram(conv_integer(wr_addr_all)) <= ram_word(7 downto 0);
					  --increment wr_addr_all because the next ram space will save the MSBs
					  wr_addr_all <= wr_addr_all + 1;
					  ram_state <= '0';
				  end case;
			 end if;
			 
        when others =>        
			if a /= "100" then --move from transmit ready to writing 
				storage_state <= '0';
				read_pause <= '1'; --output to tx_rx
			else  --transmit ready
				final_word_1 <= ram(conv_integer(out_addr));
				read_pause <= '0';
				
				--this next part might make more sense as a case statement
				--when a sample is ready, it waits a few clock cycles for threshold modules
				--to evaluate samples before updating the thresh_signal flags
				--then reevaluate "a"
				if ram_wr_en = '1' then
					delay_begin <= '1'; --delay before reevaluating "a"
				end if;
				
				if delay_begin = '1' then
					if sto_delay = "11" then
						a <= r;
						wr_addr_all <= wr_addr_sto;
						delay_begin <= '0';
						sto_delay <= "00";
					else
						sto_delay <= sto_delay + 1;
					end if;
				end if;
				
				case read_state is
						when "000" =>
							read_end <= '0'; -- move this outside of if statement?
							wr_delay <= '0';
							if read_begin = '1' then -- input from tx_rx
								--assign start and end addresses
								out_addr <= final_word_addr;
								out_addr_end <= final_word_addr_end;
								read_state <= "001";
								
							end if;
						when "001" =>
							--this is a confusing part, never fully got a handle on it
							--this can be interrupted by the UM232H(txe_lo), and by incoming samples
							--that need to be written into storage RAM
							--I had to throw in one one cycle clock delays to make the timing work
							if (txe_lo = '0' and wr_lo = '0') then
								if wr_delay = '1' then 
									out_addr <= out_addr + 1;
									--because these various signals take a couple clock cycles to propogate
									--they are triggered a couple clock cycles before the address has reached its 
									--final destination. there is probably a more elegant way to do this
									--
									if (out_addr = out_addr_end - 2) then
										read_end <= '1'; 
									elsif (out_addr = out_addr_end - 1) then
										read_state <= "000";		
										read_end <= '0';
									end if;
								else
									wr_delay <= '1';
								end if;
							end if;
						when others	=>
				end case;
			end if;
      end case;
    end if;
  end process;
end syn;