library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
library unisim;
use unisim.vcomponents.all;

--This module reads every word from ADC channel 3. If ram isn't full it compares the sample with the
--threshold level set by user.  If the threshold has been crossed, then thresh_signal_3 goes high for 256 
--word cycles (a word cycle meaning the length of time it takes for ram_wr_en to go high).
--thresh_signal_3 is sent to storage module and allows it to write into ram if it isn't full.

entity threshold_simple_3 is
  port (
    clk               : in  std_logic;
    rst               : in  std_logic;
    threshold_level_3 : in  std_logic_vector(11 downto 0);
    adc_word_final_3  : in  std_logic_vector(15 downto 0);
    ram_wr_en         : in  std_logic;   --write enable from adc
    chan_3_begin      : in  std_logic;
	 ram_full_3			 : in  std_logic;
    thresh_signal_3   : out std_logic);  --strobe when threshold is crossed

end threshold_simple_3;

architecture Behavioral of threshold_simple_3 is

  signal word_test    : std_logic_vector(11 downto 0);
  signal thresh_state : std_logic_vector(2 downto 0);
  signal pre_thresh   : std_logic_vector(3 downto 0);
  signal sample_cnt   : std_logic_vector(7 downto 0);

begin

  adc_read : process(clk, rst)

  begin

    if rst = '1' then
      thresh_state    <= "000";
      pre_thresh      <= (others => '0');
      sample_cnt      <= (others => '0');
      thresh_signal_3 <= '0';
    elsif (clk'event and clk = '1') then
      case thresh_state is
        when "000" =>  --let the ram buffer fill up
          if ram_wr_en = '1' then
            pre_thresh <= pre_thresh + 1;
            if pre_thresh = x"F" then   --16 samples load into buffer
              thresh_state <= "001";
              pre_thresh   <= (others => '0');
            end if;
          end if;
        when "001" =>
          if (ram_wr_en = '1' and ram_full_3 = '0') then  --sample word from ADC if ram is not full
            word_test    <= adc_word_final_3(11 downto 0);
            thresh_state <= "010";
          end if;
        when "010" =>
          if chan_3_begin = '1' then  --permission from tx_rx. comparisons doesn't start until threshold level is set by user
            if (word_test > threshold_level_3) then  --compare sample to threshole level
              thresh_signal_3 <= '1';
              thresh_state    <= "011";
            else
              thresh_state <= "001";  -- if threshold isn't crossed, go back to state 001 and wait for a new sample
            end if;
          end if;
        when "011" =>
          if ram_wr_en = '1' then
            sample_cnt <= sample_cnt + 1;
            if sample_cnt = x"FF" then --after 256 words, reset threshold 
              thresh_signal_3 <= '0';
              thresh_state    <= "001";
            end if;
          end if;
        when others =>
          thresh_state <= "001";
      end case;
    end if;
  end process;
end Behavioral;

