library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
library unisim;
use unisim.vcomponents.all;

--This module handles the protocol for all data coming and going
--USB module actually transmits it

entity tx_rx is
  port(
    clk, rst       : in  std_logic;
	 --control pins from UM232H
    rxf_lo, txe_lo : in  std_logic;
    --data that goes to and from USB
	 data_in        : in  std_logic_vector(7 downto 0);
    data_out       : out std_logic_vector(7 downto 0);
	 --write enable, internal control for USB module
	 wr_en          : out std_logic;
	 --signal from ADC saying that new sample is available    
	 ram_wr_en : in std_logic;
	
    --threshold level sent to threshold modules
    threshold_level_1 : out std_logic_vector(11 downto 0);
    threshold_level_2 : out std_logic_vector(11 downto 0);
    threshold_level_3 : out std_logic_vector(11 downto 0);
    threshold_level_4 : out std_logic_vector(11 downto 0);

	 --control signals to allow threshold modules to start working 
    chan_1_begin : out std_logic;
    chan_2_begin : out std_logic;
    chan_3_begin : out std_logic;
    chan_4_begin : out std_logic;
	 
	 --final word from storage module
    final_word_1    : in  std_logic_vector(7 downto 0);
    
	 --beginning and ending addresses for reading from storage module
	 final_word_addr : out std_logic_vector(15 downto 0);
	 final_word_addr_end : out std_logic_vector(15 downto 0);
	 
	 --wave count from storage module for status byte
    wave_cnt_1 : in std_logic_vector(1 downto 0);
    wave_cnt_2 : in std_logic_vector(1 downto 0);
    wave_cnt_3 : in std_logic_vector(1 downto 0);
    wave_cnt_4 : in std_logic_vector(1 downto 0);

	 --ram full flags from storage module
    ram_full_1 : in std_logic;
    ram_full_2 : in std_logic;
    ram_full_3 : in std_logic;
    ram_full_4 : in std_logic;

	 --control signals between storage and tx_rx
    read_begin : out std_logic;
	 read_pause : in std_logic;
	 read_end	: in std_logic;
	 
	 --reset for each channel of storage module
	 tx_reset_1 : out std_logic;
	 tx_reset_2 : out std_logic;
	 tx_reset_3 : out std_logic;
	 tx_reset_4 : out std_logic;
	 
	 --control signals from USB module
    wr_lo : in std_logic;
    rd_lo : in std_logic); 

end tx_rx;

architecture Behavioral of tx_rx is
  
  --array to hold incoming commands
  type reg_array is array(15 downto 0) of std_logic_vector(7 downto 0);  --register to hold incoming commands
  signal reg : reg_array;

  --counter for reading commands from USB
  signal cnt              : std_logic_vector(3 downto 0);
  
  --overall state for moving through the loop from reading the buffer, making the echo, to interpreting and executing the protocol
  signal state            : std_logic_vector(2 downto 0);
  --delay to allow pins on FPGA and UM232H to turn around 
  signal cnt_delay        : std_logic_vector(1 downto 0);
  --multiplexer selection signal
  signal mult_sel         : std_logic_vector(3 downto 0);
  --signal to control state machine for reading out a waveform from storage
  signal read_wave_state : std_logic_vector(2 downto 0);
  --signal to control state machine for reading out what commands are stored in the register: reg(1) and reg(2)
  signal reg_case        : std_logic;
  --two bytes that return which rams are full and how many waveforms are currently stored
  signal status_byte     : std_logic_vector(15 downto 0);
  --signal to control state machine for output of status through UM232H
  signal status_state    : std_logic_vector(2 downto 0);
  
  signal threshold_level_1_sig: std_logic_vector(11 downto 0);
  signal threshold_level_2_sig: std_logic_vector(11 downto 0);
  signal threshold_level_3_sig: std_logic_vector(11 downto 0);
  signal threshold_level_4_sig: std_logic_vector(11 downto 0);
  

begin
	 --output multiplexer
  with mult_sel select                  -- write from FPGA to UM232H
    data_out <=
    reg(0)        when "0000",          --echo command
    final_word_1  when "0010",
    reg(1)        when "0100",
    reg(2)        when "0101",
    status_byte(7 downto 0)   when "0110",  --status byte 1
    status_byte(15 downto 8)  when "0111",  --status byte 2
    x"ff"         when others;

  --assign signals to outputs to go to threshold modules
  threshold_level_1 <= threshold_level_1_sig;
  threshold_level_2 <= threshold_level_2_sig;
  threshold_level_3 <= threshold_level_3_sig;
  threshold_level_4 <= threshold_level_4_sig;
  
  --status byte (command 70)
  status_byte   <= ('0' & ram_full_4 & wave_cnt_4 & '0' & ram_full_3 & wave_cnt_3 & '0' & ram_full_2 & wave_cnt_2 & '0' & ram_full_1 & wave_cnt_1);
 
  tx_rx : process(clk, rst, reg)
  begin

    if rst = '1' then
      cnt              <= (others => '0');
      cnt_delay        <= (others => '0');
      state            <= (others => '0');
      wr_en            <= '0';
      mult_sel         <= "0000";
      mult_sel         <= "0000";
      read_wave_state  <= (others => '0');
      chan_1_begin     <= '0';
      chan_2_begin     <= '0';
      chan_3_begin     <= '0';
      chan_4_begin     <= '0';
		read_begin		  <= '0';
		final_word_addr  <= x"0000";
		final_word_addr_end  <= x"2000";
		threshold_level_1_sig <= (others => '0');
		threshold_level_2_sig <= (others => '0');
		threshold_level_3_sig <= (others => '0');
		threshold_level_4_sig <= (others => '0');
      status_state     <= (others => '0');
		tx_reset_1 		  <= '0';
		tx_reset_2 		  <= '0';
		tx_reset_3 		  <= '0';
		tx_reset_4 		  <= '0';
      reg_case         <= '0';

    elsif (clk'event and clk = '1') then
      case state is
        when "000" =>
		  --wait for UM232H to send commands
          if (rd_lo = '0' and rxf_lo = '0') then     -- read 
            reg(conv_integer(cnt)) <= data_in;
            cnt                    <= cnt + 1;
          elsif (rd_lo = '0' and rxf_lo = '1') then  -- reset read delay after read cycle
            cnt   <= "0000";
            state <= "001";
          end if;
        when "001" =>	--delay to allow pins time to switch directions
          if cnt_delay = x"2" then
				--once delay has reached 2, reset delay and move to next state
            state     <= "010";
            cnt_delay <= (others => '0');
          else
            cnt_delay <= cnt_delay + 1;
          end if;
          
        when "010" =>                   --Echo command
			 --enable write, send to USB module to allow output 
          wr_en    <= '1';
			 --selects reg(0), or first command in buffer, as output 
          mult_sel <= "0000";
			 state    <= "011";

			--aside from the echo, all READ and WRITE operations start here
			--commands consist of one command byte followed by parameters,
			--which change depending on the command. this part is most easily
			--explained with binary, but remember that the commands from the 
			--computer or tablet will most likely be sent using hex.
			--
			--Read operations have a '0' in the 7th bit(0 index) 
			--Write operations have a '1' in the 7th bit
			--reg(0)(7) refers to the 7th bit of the command byte
			--which is why it is used as the case statement
			--In hex, this means 0-7 are read bytes, 8-F are write bytes
			--An "x" means that any value at the moment can be used, as
			--it is currently not important
			--A "y" means that different values will do different things
			--and each case will be explained below
			
			
        when "011" =>
          case reg(0)(7) is
            when '0' =>                 -- READ operations 
              
              case reg(0)(6 downto 4) is
                when "000" =>           --reset write 0x 5a 0y
						wr_en <= '0';			 --0x is the reset command, 5a is the password
						state <= "111";		 --0y determines which channel will be reset, whatever is a "1" in binary will be reset, meaning
						if reg(1) = x"5a" then--0F will reset all channel, 03 will reset channels 1 and 2, 01 will only reset channel 1
						
							--reset channel 1 in storage if ram is full and 1st channel was selected in command code
							if (ram_full_1 = '1' and reg(2)(0) = '1') then
								tx_reset_1 <= '1';
							end if;
							
							--reset channel 2 in storage if ram is full and 1st channel was selected in command code
							if (ram_full_2 = '1' and reg(2)(1) = '1') then
								tx_reset_2 <= '1';
							end if;
							
							--reset channel 3 in storage if ram is full and 1st channel was selected in command code
							if (ram_full_3 = '1' and reg(2)(2) = '1') then
								tx_reset_3 <= '1';
							end if;
							
							--reset channel 4 in storage if ram is full and 1st channel was selected in command code
							if (ram_full_4 = '1' and reg(2)(3) = '1') then
								tx_reset_4 <= '1';
							end if;

						end if;
						
					 --Read waveform write 2x yy yy yy yy
					 --two bytes after the command are the start address
					 --followed by two bytes of the end address. range is 00 00 to 20 00
					 --Probably best to read out at least 10 or more addresses.
					 
					 when "010" =>           		
                  case read_wave_state is		
                    when "000" =>				
							--disable writing because it must first be verified that storage module is free
							--to transmit
							wr_en <= '0';				
							if read_pause = '0' then --check that storage module is not writing
								--tell storage module to start transmitting
								read_begin 		  <= '1';
								--enable writing
								wr_en 			  <= '1';
								--select output from storage
								mult_sel 		  <= "0010";
								--assign beginning and end address from registers
								final_word_addr  <= reg(1) & reg(2);
								final_word_addr_end  <= reg(3) & reg(4);
								--move to next state
								read_wave_state  <= "001";
							end if;
                    when "001" =>--everything is taken care of in storage module
						  --this module just waits for it to be finished
						  --and then controls the USB module
							read_begin <= '0';
							--if the end of the address range has been reached, disable writing
							if read_end = '1' then
								read_wave_state <= "010";
								wr_en <= '0';
							else
								--if the writing to UM232H is interrupted by the UM232H filling
								--up its buffer or by an incoming sample being written into 
								--storage ram, this next part disables writing to the UM232H
								--and then enables it as soon as the interruption has passed
								if (read_pause = '0' and txe_lo = '0') then
									wr_en <= '1';
								else
									wr_en <= '0';
								end if;	
							end if;
						  
						  when others =>
							state <= "111";
							read_wave_state <= "000";
						  
                  end case;
						
						--Read input command register 3x
						--This simple returns the two value currently
						--in the array directly after the command byte
						--reg(1) and reg(2)
                when "011" =>           
                  case reg_case is
                    when '0' =>
                      mult_sel <= "0100";
                      reg_case <= '1';
                    when others =>
                      mult_sel <= "0101";
                      wr_en    <= '0';
                      reg_case <= '0';
                      state    <= "111";
                  end case;
                  
						
					--Status byte 7x	
					--whatever you get back must be converted into 16 bits of binary. Each grouping of four bits gives 
					--information for that channel. In each grouping, the two least significant bits are reserved for 
					--future use to count waves stored in that channel. The third bit, which you want to pay attention to, 
					--siginfies whether or not the ram is full. The 4th bit in each group is always zero. So the binary number 
					--0000000001000000 means that the 2nd channel is full. 0000000000000100 means the first channel is full.	
					
                when "111" =>           
                  case status_state is
                    when "000" => --delay
                      status_state <= status_state + 1;
                    when "001" => --status byte 1
							 mult_sel     <= "0110";
                      status_state <= status_state + 1;
                    when others => --status byte 2
						    mult_sel     <= "0111";
                      wr_en        <= '0';
							 status_state <= "000";
                      state        <= "111";
                  end case;
                  
                when others =>
                  state <= "111";
              end case;
              
            when others =>              -- WRITE operations
				
					--Threshold set
					--The threshold modules don't start working until a threshold is set
					--which is what the "chan_1_begin" type signals mean
					--This section also sets the threshold level
					--0y refers to the channel. 01 is channel 1, 02 is channel 2...05 is all channels
					--yy yy  is the actual threshold, it is a 12 bit number, so the MSB 4 bits are ignored
					--Min threshold is 00 00, Max threshold is 0F FE. 0F FF should not be used because
					--it is impossible for this threshold to be crossed.
					
              case reg(0)(6 downto 4) is
                when "010" =>           -- Wait for trigger  write Ax 0y xy yy
                  wr_en <= '0';
                  state <= "111";
                  if reg(1) = x"01" then --set threshold level and threshold module enable for channel 1
                    threshold_level_1_sig <= (reg(2)(3 downto 0) & reg(3));
                    chan_1_begin      <= '1';
                  elsif reg(1) = x"02" then --set threshold level and threshold module enable for channel 2
                    threshold_level_2_sig <= (reg(2)(3 downto 0) & reg(3));
                    chan_2_begin      <= '1';
                  elsif reg(1) = x"03" then --set threshold level and threshold module enable for channel 3
                    threshold_level_3_sig <= (reg(2)(3 downto 0) & reg(3));
                    chan_3_begin      <= '1';
                  elsif reg(1) = x"04" then --set threshold level and threshold module enable for channel 4
						threshold_level_4_sig <= (reg(2)(3 downto 0) & reg(3));
                    chan_4_begin      <= '1';
                  elsif reg(1) = x"05" then --this command sets all threshold levels and enables all threshold modules
                    chan_1_begin      <= '1';
                    chan_2_begin      <= '1';
                    chan_3_begin      <= '1';
                    chan_4_begin      <= '1';
                    threshold_level_1_sig <= (reg(2)(3 downto 0) & reg(3));
                    threshold_level_2_sig <= (reg(2)(3 downto 0) & reg(3));
                    threshold_level_3_sig <= (reg(2)(3 downto 0) & reg(3));
                    threshold_level_4_sig <= (reg(2)(3 downto 0) & reg(3));
                  end if;
                when others =>
                  state <= "111";
              end case;
          end case;

        when others =>
          wr_en <= '0';
          state <= "000";
			 tx_reset_1 <= '0';
			 tx_reset_2 <= '0';
			 tx_reset_3 <= '0';
			 tx_reset_4 <= '0';
      end case;
    end if;
  end process;
end Behavioral;
