library IEEE;
use IEEE.STD_LOGIC_1164.all;
use IEEE.STD_LOGIC_UNSIGNED.all;
library unisim;
use unisim.vcomponents.all;

--This module communicates directly with the UM232H, which is active low. Information is bidirectional
--going to and from the device over the same set of pins, so the data in/data out section on the bottom 
--is a tri state buffer. For timing diagram and explanation, refer to datasheet for FT232H and find 
--FT245 Synchronous FIFO mode.
-- http://www.ftdichip.com/Support/Documents/DataSheets/ICs/DS_FT232H.pdf pg 27-28

entity usb is
  port (
    clk, rst            : in    std_logic;
    rxf_lo, txe_lo      : in    std_logic;  --signals coming from UM232H
    data                : inout std_logic_vector(7 downto 0);  -- connect to device data
    data_in             : out   std_logic_vector(7 downto 0);  -- input data
    data_out            : in    std_logic_vector(7 downto 0);  -- output data
    wr_en               : in    std_logic;  --control signal coming from module tx_rx
    oe_lo, rd_lo, wr_lo : out   std_logic); --signals going to UM232H

end usb;

architecture Behavioral of usb is

  signal oe_ready : std_logic;
  signal dir      : std_logic;

begin

  read_write : process(clk, rst, oe_ready, rxf_lo, txe_lo, dir)
  begin
    
    if rst = '1' then
      oe_lo    <= '1';
      rd_lo    <= '1';
      wr_lo    <= '1';
      dir      <= '1';
      oe_ready <= '0';
    elsif (clk'event and clk = '1') then
      if rxf_lo = '0' then                      --read cycle
        if oe_ready = '1' then --oe_lo goes low first, then rd_lo goes low 
          rd_lo <= '0';
        else 
          oe_lo    <= '0';
          oe_ready <= '1';
        end if;
      else
        oe_ready <= '0';
        oe_lo    <= '1';
        rd_lo    <= '1';
        if (txe_lo = '0' and wr_en = '1') then  --write cycle. txe_lo comes from the 
			 wr_lo <= '0';									--UM232H, wr_en comes from the tx_rx 
          dir   <= '0'; -- dir controls tristate buffer
        else
          wr_lo <= '1';
          dir   <= '1';
        end if;
      end if;
    end if;
  end process;

  data_in <= data;
  with dir select                       -- write from FPGA to UM232H, tri state buffer
    data <=										
    "ZZZZZZZZ" when '1',
    data_out   when others;
end Behavioral;
